from collections import namedtuple
from pathlib import Path

import pytest
import responses

from openhexa.toolbox.dhis2 import Api

VERSIONS = ["2.36", "2.37", "2.38", "2.39", "2.40", "2.41"]


@pytest.fixture
def con():
    Connection = namedtuple("Connection", ["url", "username", "password"])
    return Connection("http://localhost:8080", "admin", "district")


@responses.activate
@pytest.mark.parametrize("version", VERSIONS)
def test_authenticate(version, con):
    responses_dir = Path("tests", "dhis2", "responses", version)
    responses._add_from_file(Path(responses_dir, "api_authenticate.yaml"))
    Api(con, cache_dir=None)


@responses.activate
@pytest.mark.parametrize("version", VERSIONS)
def test_get(version, con):
    responses_dir = Path("tests", "dhis2", "responses", version)
    responses._add_from_file(Path(responses_dir, "api_authenticate.yaml"))
    responses._add_from_file(Path(responses_dir, "api_get.yaml"))
    api = Api(con, cache_dir=None)
    r = api.get("system/info")
    assert "version" in r


@responses.activate
@pytest.mark.parametrize("version", VERSIONS)
def test_get_paged(version, con):
    responses_dir = Path("tests", "dhis2", "responses", version)
    responses._add_from_file(Path(responses_dir, "api_authenticate.yaml"))
    responses._add_from_file(Path(responses_dir, "api_get_paged.yaml"))
    api = Api(con, cache_dir=None)
    r = api.get_paged(endpoint="organisationUnits", params={"fields": "id,name"})
    for page in r:
        assert page



iaso_mocked_auth_token = {
    "access": "eyJhbGciOiJIUzI1NiJ9.eyJSb2xlIjoiQWRtaW4iLCJJc3N1ZXIiOiJJc3N1ZXIiLCJVc2VybmFtZSI6IkphdmFJblVzZSIsImV4cCI6MTcxNzY5MDEwNCwiaWF0IjoxNzE3NjkwMTA0fQ.WsmnKvyKFR2eWNL4wD4yrnd6F9CDBV2dCaMx9lE6V84",  # noqa: E501
    "refresh": "eyJhbGciOiJIUzI1NiJ9.eyJSb2xlIjoiQWRtaW4iLCJJc3N1ZXIiOiJJc3N1ZXIiLCJVc2VybmFtZSI6IkphdmFJblVzZSIsImV4cCI6MTcxNzY5MDEwNCwiaWF0IjoxNzE3NjkwMTA0fQ.WsmnKvyKFR2eWNL4wD4yrnd6F9CDBV2dCaMx9lE6V84",  # noqa: E501
}
iaso_mocked_refreshed_auth_token = {
    "access": "eyJhbGciOiJIUzI1NiJ9.eyJSb2xlIjoiQWRtaW4iLCJJc3N1ZXIiOiJJc3N1ZXIiLCJVc2VybmFtZSI6IkphdmFJblVzZSIsImV4cCI6MTcxNzY5MDEwNCwiaWF0IjoxNzE3NzYwMTA0fQ._pXcqDw0QgvznvNuhVPwYyIms3H5imH-q6A7lIQJjYQ",  # noqa: E501
    "refresh": "eyJhbGciOiJIUzI1NiJ9.eyJSb2xlIjoiQWRtaW4iLCJJc3N1ZXIiOiJJc3N1ZXIiLCJVc2VybmFtZSI6IkphdmFJblVzZSIsImV4cCI6MTcxNzY5MDEwNCwiaWF0IjoxNzE3NjkwMTA0fQ.WsmnKvyKFR2eWNL4wD4yrnd6F9CDBV2dCaMx9lE6V84",  # noqa: E501
}


iaso_mocked_forms = {
    "forms": [
        {
            "id": 278,
            "name": "Test (form styling)",
            "form_id": "pathways_indonesia_survey_1",
            "device_field": "deviceid",
            "location_field": "",
            "org_unit_types": [{"id": 781, "name": "Province", "short_name": "Prov", "created_at": 1712825023.047433}],
            "projects": [{"id": 149, "name": "Pathways"}],
            "created_at": 1713171086.141424,
        }
    ]
}

iaso_mocked_projects = {
    "projects": [
        {
            "id": 149,
            "name": "Pathways",
            "app_id": "pathways",
            "feature_flags": [
                {"id": 3, "name": "GPS point for each form", "code": "TAKE_GPS_ON_FORM"},
                {"id": 7, "name": "Mobile: Show data collection screen", "code": "DATA_COLLECTION"},
                {"id": 12, "name": "Mobile: Finalized forms are read only", "code": "MOBILE_FINALIZED_FORM_ARE_READ"},
                {"id": 4, "name": "Authentication", "code": "REQUIRE_AUTHENTICATION"},
            ],
            "created_at": 1710153966.532745,
            "updated_at": 1717664805.185712,
            "needs_authentication": True,
        }
    ]
}

iaso_mocked_orgunits = {
    "orgUnits": [
        {
            "name": "ACEH",
            "id": 1978297,
            "parent_id": 1978331,
            "org_unit_type_id": 781,
            "org_unit_type_name": "Province",
            "validation_status": "VALID",
            "created_at": 1712825023.085615,
            "updated_at": 1712828860.665764,
        }
    ]
}

iaso_mocked_instances = {
    "count": 28,
    "instances": [
        {
            "uuid": "4dc8c051-5b86-4b8e-b767-137e017d3c07",
            "export_id": "ArtNWdBsfAj",
            "file_name": "276_513ae6cf-5efa-4754-b7fb-2d2124ef3efd_2024-04-11_14-17-05.xml",
            "file_content": {
                "age": "14",
                "name": "qfdsf",
                "U1_u1": "U1",
                "address": "qsdfqsdf",
                "endtime": "2024-04-11T14:18:46.435+02:00",
                "village": "dddddd",
                "_version": "6004241215",
                "deviceid": "853e64340cfa95b8",
                "district": "badung",
                "location": "urban",
                "religion": "protestant",
                "username": "",
                "ethnicity": "sasak",
                "pregnancy": "yes",
                "starttime": "2024-04-11T14:17:05.394+02:00",
                "time_slot": "14_to_16",
                "instanceID": "uuid:676558f6-17f2-4a4f-a8f0-08b43cc572da",
                "occupation": "ffffffff",
                "subdistrict": "dddd",
                "children_age": "12,12",
                "partner_name": "",
                "phone_number": "111",
                "type_of_work": "formal",
                "have_children": "yes",
                "type_of_phone": "feature_phone",
                "devicephonenum": "",
                "final_comments": "thanks",
                "jkn_mobile_app": "no",
                "marital_status": "single",
                "preferred_date": "2024-04-11",
                "access_to_phone": "yes",
                "number_of_children": "1",
                "type_jnk_insurance": "pbi",
                "living_with_partner": "no",
                "enrolled_jnk_insurance": "yes",
                "person_with_disability": "yes",
                "education_participant_1": "no_education",
                "screening_questions_intro": "",
                "preferred_language_interview": "bahasa_daerah",
            },
            "file_url": "https://iaso-stg.s3.amazonaws.com/instances/276_513ae6cf-5efa-4754-b7fb-2d2124ef3efd_2024-04-11_14-17-05.xml?X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Credential=AKIAS4KZU3S6DZBAV5GW%2F20240611%2Feu-central-1%2Fs3%2Faws4_request&X-Amz-Date=20240611T075045Z&X-Amz-Expires=3600&X-Amz-SignedHeaders=host&X-Amz-Signature=7ebdbe2c44d8251589b31596aa61cf488fae5970a64a76084804b7c7cb9356c7",
            "id": 36627,
            "form_id": 276,
            "form_name": "Pathways Indonesia Recruitment (test)",
            "created_at": 1712837926.0,
            "updated_at": 1712837933.562898,
            "org_unit": {
                "name": "ACEH",
                "short_name": "ACEH",
                "id": 1978297,
                "source": "Indonesia",
                "source_id": 192,
                "source_ref": "ID001001000000000000",
                "parent_id": 1978331,
                "org_unit_type_id": 781,
                "org_unit_type_name": "Province",
                "org_unit_type_depth": 1,
                "created_at": 1712825023.085615,
                "updated_at": 1712828860.665764,
                "aliases": None,
                "validation_status": "VALID",
                "latitude": None,
                "longitude": None,
                "altitude": None,
                "has_geo_json": True,
                "version": 0,
                "opening_date": None,
                "closed_date": None,
            },
            "latitude": 50.8267504,
            "longitude": 4.3505693,
            "altitude": 113.30000305175781,
            "period": None,
            "status": "READY",
            "correlation_id": 36627503,
            "created_by": {"username": "mdewulf-pathways", "first_name": "Martin", "last_name": "De Wulf"},
            "last_modified_by": "mdewulf-pathways",
            "can_user_modify": True,
            "is_locked": False,
            "is_instance_of_reference_form": False,
            "is_reference_instance": False,
        },
        {
            "uuid": "d2fc1f2f-ed83-47db-b4db-be2679a26aff",
            "export_id": "pNI8aE6Am2Z",
            "file_name": "276_1770f9ff-254f-43af-9c13-21679a12a3d6_2024-04-12_11-47-29.xml",
            "file_content": {
                "age": "30",
                "name": "Test 1",
                "R1_r1": "R1",
                "address": "Test 1",
                "endtime": "2024-04-12T11:51:29.661+03:00",
                "village": "Test 1",
                "_version": "6004241215",
                "deviceid": "cb5f71d5c9b0f248",
                "district": "pidie",
                "location": "rural",
                "religion": "protestant",
                "username": "",
                "ethnicity": "aceh",
                "pregnancy": "no",
                "starttime": "2024-04-12T11:47:29.176+03:00",
                "time_slot": "12_to_14",
                "instanceID": "uuid:3b07bced-1155-4a3a-9324-189653cac25e",
                "occupation": "Test 1",
                "subdistrict": "Test 1",
                "partner_name": "",
                "phone_number": "235866358",
                "type_of_work": "informal",
                "have_children": "no",
                "type_of_phone": "smartphone",
                "devicephonenum": "",
                "final_comments": "Test 1",
                "jkn_mobile_app": "yes",
                "marital_status": "married_partner",
                "preferred_date": "2024-04-12",
                "access_to_phone": "yes",
                "jkn_mobile_app_use": "sometimes_use_it",
                "type_jnk_insurance": "pbi",
                "living_with_partner": "no",
                "enrolled_jnk_insurance": "yes",
                "person_with_disability": "no",
                "education_participant_1": "some_education",
                "hh_member_with_disability": "no",
                "screening_questions_intro": "",
                "preferred_language_interview": "bahasa_indonesia",
            },
            "file_url": "https://iaso-stg.s3.amazonaws.com/instances/276_1770f9ff-254f-43af-9c13-21679a12a3d6_2024-04-12_11-47-29.xml?X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Credential=AKIAS4KZU3S6DZBAV5GW%2F20240611%2Feu-central-1%2Fs3%2Faws4_request&X-Amz-Date=20240611T075045Z&X-Amz-Expires=3600&X-Amz-SignedHeaders=host&X-Amz-Signature=05eb2a925b34dd620cf35b27a0b9850f0dc4f0351e96a4c41ad4d82da58f11e6",
            "id": 36628,
            "form_id": 276,
            "form_name": "Pathways Indonesia Recruitment (test)",
            "created_at": 1712911889.0,
            "updated_at": 1712911890.646247,
            "org_unit": {
                "name": "ACEH",
                "short_name": "ACEH",
                "id": 1978297,
                "source": "Indonesia",
                "source_id": 192,
                "source_ref": "ID001001000000000000",
                "parent_id": 1978331,
                "org_unit_type_id": 781,
                "org_unit_type_name": "Province",
                "org_unit_type_depth": 1,
                "created_at": 1712825023.085615,
                "updated_at": 1712828860.665764,
                "aliases": None,
                "validation_status": "VALID",
                "latitude": None,
                "longitude": None,
                "altitude": None,
                "has_geo_json": True,
                "version": 0,
                "opening_date": None,
                "closed_date": None,
            },
            "latitude": 60.1916042,
            "longitude": 24.9458922,
            "altitude": 46.978047416201505,
            "period": None,
            "status": "READY",
            "correlation_id": 36628816,
            "created_by": None,
            "last_modified_by": None,
            "can_user_modify": True,
            "is_locked": False,
            "is_instance_of_reference_form": False,
            "is_reference_instance": False,
        },
    ],
    "has_next": True,
    "has_previous": False,
    "page": 1,
    "pages": 14,
    "limit": 2,
}



import json
import logging
import zlib
from collections import OrderedDict
from pathlib import Path
from typing import Iterable, Optional, Protocol, Sequence, Union
from urllib.parse import urlparse

import requests
from diskcache import DEFAULT_SETTINGS, Cache
from requests.adapters import HTTPAdapter
from urllib3 import Retry

logger = logging.getLogger(__name__)


class DHIS2Error(Exception):
    pass


class DHIS2Connection(Protocol):
    username: str
    password: str
    url: str


class Api:
    def __init__(self, connection: DHIS2Connection, cache_dir: Optional[Union[Path, str]] = None):
        self.url = self.parse_api_url(connection.url)

        self.session = requests.Session()
        adapter = HTTPAdapter(
            max_retries=Retry(
                total=3,
                backoff_factor=5,
                allowed_methods=["HEAD", "GET"],
                status_forcelist=[429, 500, 502, 503, 504],
            )
        )
        self.session.mount("https://", adapter)
        self.session.mount("http://", adapter)

        self.session = self.authenticate(connection.username, connection.password)

        self.cache = None
        if cache_dir:
            self.cache = ApiCache(cache_dir, self.url)

        self.PAGE_SIZE = 1000

        self.DEFAULT_EXPIRE_TIME = 86400
        self.EXPIRE_TIMES = {
            "dataValueSets": 604800,
            "analytics": 604800,
            "system": 60,
        }

    @staticmethod
    def parse_api_url(url: str) -> str:
        """Ensure that API URL is correctly formatted."""
        url = url.rstrip("/")
        if "/api" not in url:
            url += "/api"
        return url

    @staticmethod
    def raise_if_error(response: requests.Response):
        """Raise DHIS2Error with message provided by API."""
        # raise DHIS2 error if error message is provided
        if response.status_code != 200 and "json" in response.headers["content-type"]:
            msg = response.json()
            if msg.get("status") == "ERROR":
                raise DHIS2Error(f"{msg.get('status')} {msg.get('httpStatusCode')}: {msg.get('message')}")

        # raise with requests if no error message provided
        response.raise_for_status()

    def authenticate(self, username: str, password: str) -> requests.Session:
        """Authentify using Basic Authentication."""
        s = requests.Session()
        s.auth = requests.auth.HTTPBasicAuth(username, password)
        r = s.get(f"{self.url}/system/ping")
        self.raise_if_error(r)
        logger.info(f"Logged in to '{self.url}' as '{username}'")
        return s

    def get(self, endpoint: str, params: dict = None, use_cache: bool = True) -> dict:
        """Send GET request and return JSON response as a dict."""
        r = requests.Request(method="GET", url=f"{self.url}/{endpoint}", params=params)
        url = r.prepare().url
        logger.debug(f"API request {url}")

        use_cache = self.cache and use_cache

        if use_cache:
            r = self.cache.get(endpoint=endpoint, params=params)
            if r:
                return r

        r = self.session.get(f"{self.url}/{endpoint}", params=params)
        self.raise_if_error(r)

        if use_cache:
            self.cache.set(endpoint=endpoint, params=params, response=r.json())

        return r.json()

    def get_paged(self, endpoint: str, params: dict = None, use_cache: bool = True) -> Iterable[requests.Response]:
        """Iterate over paged responses."""
        use_cache = self.cache and use_cache

        if not params:
            params = {}
        params["pageSize"] = self.PAGE_SIZE

        # 1st page
        r = self.get(endpoint=endpoint, params=params, use_cache=use_cache)
        yield r

        if "pager" in r:
            params["page"] = r["pager"]["page"]
            while "nextPage" in r["pager"]:
                params["page"] += 1
                r = self.get(endpoint=endpoint, params=params, use_cache=use_cache)
                yield r

    @staticmethod
    def merge_pages(pages: Sequence[requests.Response]) -> dict:
        """Merge lists from paged responses.

        The "pager" key in the response will be removed and all keys of type `list`
        (e.g. organisationUnits) will be merged into a single one.

        Parameters
        ----------
        pages : list of responses
            A list of paged responses, as returned by Api.get_paged()

        Return
        ------
        dict
            Merged response as a dict with merged lists
        """
        merged_response = {}
        first_page = pages[0]
        for key in first_page.keys():
            if isinstance(first_page[key], list):
                merged_response[key] = []
                for page in pages:
                    merged_response[key] += page[key]
        return merged_response

    def post(self, endpoint: str, json: dict = None, params: dict = None) -> requests.Response:
        r = self.session.post(f"{self.url}/{endpoint}", json=json, params=params)
        self.raise_if_error(r)
        return r


class ApiCache:
    def __init__(self, cache_dir: Path, api_url: str):
        """Cache API requests with diskcache."""
        self.dir = cache_dir / urlparse(api_url).netloc
        self.setup()
        self.api_url = api_url

        self.DEFAULT_EXPIRE_TIME = 86400
        self.EXPIRE_TIMES = {"dataValueSets": 604800, "analytics": 604800, "system": 60}
        self.SETTINGS = DEFAULT_SETTINGS
        self.SETTINGS["size_limit"] = 10000000000

    def setup(self):
        """Setup cache directory.

        Diskcache directory will be a subdir in the provided cache_dir,
        named after the DHIS2 instance domain name.
        """
        self.dir.mkdir(parents=True, exist_ok=True)
        logger.debug(f"Using cache directory {self.dir.absolute().as_posix()}")

    def expire(self):
        """Remove expired items from cache."""
        with Cache(self.dir, **self.SETTINGS) as cache:
            cache.expire()

    def clear(self):
        """Remove all items from cache."""
        with Cache(self.dir, **self.SETTINGS) as cache:
            cache.clear()

    def get_key(self, endpoint: str, params: Optional[dict]) -> str:
        """Generate cache key from API endpoint and query parameters."""
        if params:
            params = OrderedDict(sorted(params.items()))
            return f"{self.api_url}/{endpoint}/{json.dumps(params)}"
        else:
            return f"{self.api_url}/{endpoint}"

    def get(self, endpoint: str, params: Optional[dict]) -> Union[dict, None]:
        """Get JSON query response from cache as a dict."""
        key = self.get_key(endpoint=endpoint, params=params)
        with Cache(self.dir, **self.SETTINGS) as cache:
            content = cache.get(key)
            if content:
                logger.debug("Cache hit, returning decompressed response")
                return json.loads(zlib.decompress(content).decode())
        return None

    def set(self, endpoint: str, response: dict, params: Optional[dict]):
        """Cache JSON query response."""
        key = self.get_key(endpoint=endpoint, params=params)
        with Cache(self.dir, **self.SETTINGS) as cache:
            value = zlib.compress(json.dumps(response).encode())
            cache.set(key, value, expire=self.EXPIRE_TIMES.get("endpoint", self.DEFAULT_EXPIRE_TIME), retry=True)



import pytest

from unittest import mock
from uuid import uuid4

from openhexa.toolbox.hexa import OpenHEXAClient, OpenHEXA, NotFound
from .fixtures.openhexa_api_fixtures import (
    openhexa_mocked_workspaces,
    openhexa_mocked_workspaces_pipelines,
    openhexa_mocked_pipeline,
    openhexa_mocked_pipeline_run,
)


class TestOpenHEXAClient:
    def _mock_response(self, data):
        mock_response = mock.MagicMock()
        mock_response.json.return_value = data
        return mock_response

    def test_authenticate_with_creds_success(self):
        hexa = OpenHEXAClient("https://app.demo.openhexa.org")
        mock_response = self._mock_response(
            {
                "data": {
                    "login": {
                        "success": True,
                    }
                }
            },
        )

        with mock.patch.object(hexa, "_graphql_request", return_value=mock_response):
            assert hexa.authenticate(with_credentials=("username", "password")) is True

    def test_authenticate_with_creds_failed(self):
        hexa = OpenHEXAClient("https://app.demo.openhexa.org")
        mock_response = self._mock_response(
            {"data": {"login": {"success": False, "errors": ["INVALID_CREDENTIALS"]}}},
        )

        with mock.patch.object(hexa, "_graphql_request", return_value=mock_response):
            with pytest.raises(Exception) as e:
                hexa.authenticate(with_credentials=("username", "password"))
                assert str(e) == "Login failed : ['INVALID_CREDENTIALSS']"

    def test_authenticate_with_otq_enabled_failed(self):
        hexa = OpenHEXAClient("https://app.demo.openhexa.org")
        mock_response = self._mock_response(
            {"data": {"login": {"success": False, "errors": ["OTP_REQUIRED"]}}},
        )
        with mock.patch.object(hexa, "_graphql_request", return_value=mock_response):
            with pytest.raises(Exception) as e:
                hexa.authenticate(with_credentials=("username", "password"))
                assert str(e) == "Login failed : you need to disable two-factor authentication."


@mock.patch("openhexa.toolbox.hexa.hexa.OpenHEXAClient")
class TestOpenHEXA:
    def test_get_workspaces(self, mock_hexa_client):
        hexa = OpenHEXA("http://localhost:3000", token="token")

        with mock.patch.object(hexa, "query", return_value=openhexa_mocked_workspaces):
            assert hexa.get_workspaces() == openhexa_mocked_workspaces

    def test_get_pipelines(self, mock_hexa_client):
        hexa = OpenHEXA("http://localhost:3000", token="token")

        with mock.patch.object(hexa, "query", return_value=openhexa_mocked_workspaces_pipelines):
            assert hexa.get_pipelines("slug") == openhexa_mocked_workspaces_pipelines

    def test_get_pipeline_not_found(self, mock_hexa_client):
        hexa = OpenHEXA("http://localhost:3000", token="token")

        with mock.patch.object(hexa, "query", return_value=None):
            with pytest.raises(NotFound):
                hexa.get_pipeline("workspace", "code")

    def test_get_pipeline(self, mock_hexa_client):
        hexa = OpenHEXA("http://localhost:3000", token="token")

        with mock.patch.object(hexa, "query", return_value=openhexa_mocked_pipeline):
            assert hexa.get_pipeline("workspace", "code") == openhexa_mocked_pipeline

    def test_get_pipeline_run_not_found(self, mock_hexa_client):
        hexa = OpenHEXA("http://localhost:3000", token="token")

        with mock.patch.object(hexa, "query", return_value={"pipelineRun": None}):
            with pytest.raises(NotFound):
                hexa.get_pipeline_run(uuid4())

    def test_get_pipeline_run(self, mock_hexa_client):
        hexa = OpenHEXA("http://localhost:3000", token="token")

        with mock.patch.object(hexa, "query", return_value=openhexa_mocked_pipeline_run):
            assert hexa.get_pipeline_run(uuid4()) == openhexa_mocked_pipeline_run["pipelineRun"]

    def test_run_pipeline_not_found(self, mock_hexa_client):
        hexa = OpenHEXA("http://localhost:3000", token="token")
        data = {"runPipeline": {"success": False, "errors": ["PIPELINE_NOT_FOUND"]}}
        with mock.patch.object(hexa, "query", return_value=data):
            with pytest.raises(NotFound):
                hexa.run_pipeline(uuid4())

    def test_run_pipeline(self, mock_hexa_client):
        hexa = OpenHEXA("http://localhost:3000", token="token")
        run_id = uuid4()
        data = {"runPipeline": {"success": True, "errors": [], "run": {"id": run_id}}}

        with mock.patch.object(hexa, "query", return_value=data):
            assert hexa.run_pipeline(uuid4()) == {"id": run_id}



import os

import geopandas as gpd
import polars as pl
from shapely.geometry import Point

from .api import Api, Field, Survey
from .parse import cast_values


def field_from_name(name: str, survey: Survey) -> Field:
    """Get field in survey from its name."""
    for field in survey.fields:
        if field.name == name:
            return field
    return None


def lists_to_string(df: pl.DataFrame) -> pl.DataFrame:
    """Convert lists in a dataframe into comma-separated strings."""
    for column in df.columns:
        if df[column].dtype == pl.List(pl.Utf8):
            df = df.with_columns(pl.col(column).list.join(", ").alias(column))
    return df


def _download(url: str, dst_dir: str, api: Api):
    with api.session.get(url, stream=True) as r:
        fname = r.headers["Content-Disposition"].split("filename=")[-1]
        fpath = os.path.join(dst_dir, fname)
        if os.path.exists(fpath):
            return
        os.makedirs(dst_dir, exist_ok=True)
        with open(fpath, "wb") as f:
            for chunk in r.iter_content(chunk_size=1024**2):
                if chunk:
                    f.write(chunk)


def download_attachments(df: pl.DataFrame, dst_dir: str, api: Api):
    """Download all survey attachments referenced in a dataframe."""
    for row in df.iter_rows(named=True):
        attachments = row.get("_attachments")
        for attachment in attachments:
            url = attachment.get("download_url")
            if url:
                _download(url, dst_dir, api)


def rename_columns(df: pl.DataFrame) -> pl.DataFrame:
    """Use field names instead of xpaths as columns.

    E.g. "group1/DATE" becomes "DATE".
    """
    mapping = {}
    for column in df.columns:
        if "/" in column:
            mapping[column] = column.split("/")[-1]
    df = df.rename(mapping)
    return df


def to_dataframe(survey: Survey) -> pl.DataFrame:
    """Get survey data as a polars dataframe."""
    rows = survey.get_data()
    df = pl.DataFrame(rows)
    df = rename_columns(df)  # use kobo field names
    df = cast_values(df, survey)  # use kobo field types
    return df


def to_geodataframe(df: pl.DataFrame) -> gpd.GeoDataFrame:
    """Get geodataframe from formatted survey dataframe."""
    geodf = gpd.GeoDataFrame(
        data=df.to_pandas(),
        crs="EPSG:4326",
        geometry=[Point(coords[1], coords[0]) if all(coords) else None for coords in df["_geolocation"]],
    )
    geodf = geodf.drop(columns=["_geolocation"])
    return geodf


def get_fields_mapping(survey: Survey) -> pl.DataFrame:
    """Get a mapping of fields names, types and labels as a dataframe."""
    mapping = []
    for f in survey.fields:
        if f.label and f.name and f.type:
            mapping.append({"name": f.name, "type": f.type, "label": f.label})
    return pl.DataFrame(mapping)



from requests import Session
import typing


class NotFound(Exception):
    """Errors related to an element not found."""

    pass


class OpenHEXAClient:
    def __init__(self, base_url):
        self.url = base_url.rstrip("/")
        self.session = Session()
        self.session.headers.update({"Content-Type": "application/json", "User-Agent": "OpenHEXA Python Client"})

    def authenticate(
        self,
        with_credentials: typing.Optional[tuple[str, str]] = None,
        with_token: typing.Optional[str] = None,
    ):
        """
        with_credentials: tuple of email and password
        with_token: JWT token
        """
        if with_credentials:
            resp = self._graphql_request(
                """
                mutation Login($input: LoginInput!) {
                    login(input: $input) {
                        success
                        errors
                    }
                }
            """,
                {
                    "input": {
                        "email": with_credentials[0],
                        "password": with_credentials[1],
                    }
                },
            )
            resp.raise_for_status()
            data = resp.json()["data"]

            if not data["login"]["success"]:
                if "OTP_REQUIRED" in data["login"]["errors"]:
                    raise Exception("Login failed : two-factor authentication needs to be disabled.")

                raise Exception(f"Login failed : {data['login']['errors']}")
            else:
                self.session.headers["Cookie"] = resp.headers["Set-Cookie"]

        elif with_token:
            self.session.headers.update({"Authorization": f"Bearer {with_token}"})

        return True

    def _graphql_request(self, operation, variables=None):
        return self.session.post(f"{self.url}/graphql", json={"query": operation, "variables": variables})

    def query(self, operation, variables=None):
        resp = self._graphql_request(operation, variables)
        if resp.status_code == 400:
            raise Exception(resp.json()["errors"][0]["message"])
        resp.raise_for_status()
        payload = resp.json()
        if payload.get("errors"):
            raise Exception(payload["errors"])
        return payload["data"]



from .api import OpenHEXAClient, NotFound
import typing
import uuid


class OpenHEXA:
    def __init__(
        self,
        server_url: str,
        username: typing.Optional[str] = None,
        password: typing.Optional[str] = None,
        token: typing.Optional[str] = None,
    ):
        """
        Initializes the OpenHEXA client. If username and password are provided we will try to
        authenticate via credentials. It support also the authentication via pipeline token.
        Note that two-factor authentication needs to be disabled, otherwise the authentication will fail.

        Parameters
        ----------
        server_url: OpenHEXA server URL
        username: OpenHEXA instance username
        password: OpenHEXA instance password
        token: OpenHEXA pipeline token

        Raises
        ------
        Exception
            when a login or authentication error

        Examples:
            >>> from openhexa.toolbox.hexa import OpenHEXA
            >>> hexa = OpenHEXA(server_url="https://app.demo.openhexa.org",
            >>>             username="user",
            >>>             password="pass")

            or

            >>> from openhexa.toolbox.hexa import OpenHEXA
            >>> hexa = OpenHEXA(server_url="https://app.demo.openhexa.org",token="token")
        """

        self.client = OpenHEXAClient(server_url)
        if username and password:
            self.client.authenticate(with_credentials=(username, password))
        elif token:
            self.client.authenticate(with_token=token)
        else:
            raise Exception("Missing authentication credentials: username, password or token missing.")

    def query(self, operation, variables=None):
        """
        Executes a query operation using the client's API.

        This method sends a query request to the server using the specified operation and optional variables.
        The `operation` parameter defines the operation to be performed, typically in the form of a query or mutation.
        The `variables` parameter allows you to pass any required variables for the query.

        Parameters:
        ----------
        operation : str
            A GraphQL string representing the query or mutation operation to be executed.

        variables : dict | None, optional
            A dictionary of parameters to be used in the query / mutation operation.

        Returns:
        -------
        dict
            The query result from OpenHEXA server.

        Example Usage:
        --------------
        ```
        # Simple query without variables
        response = hexa.query(operation=\"""{ users { id name } }\""")

        # Query with variables
        operation = \"""
            mutation($input: RunPipelineInput!) {
            runPipeline(input: $input) {
                success
                errors
                run {
                    id
                }
            }
        }
        \"""
        variables = {"input": {"id": pipeline_id, "config": {}}}
        response = hexa.query(operation=operation, variables=variables)
        ```
        """
        return self.client.query(operation, variables)

    def get_workspaces(self, page: int = 1, per_page: int = 10) -> dict:
        """
        Fetches a paginated list of workspaces.


        Parameters:
        ----------
        page : int, optional
            The page number to retrieve. Defaults to 1.
        per_page : int, optional
            The number of items to retrieve per page. Defaults to 10.

        Returns:
        -------
        dict
            A dictionary containing the list of workspaces (`items`), the total number of items (`totalItems`),
            and the total number of pages (`totalPages`).
        """
        return self.query(
            """
           query($page: Int!, $perPage: Int!) {
              workspaces(page: $page, perPage: $perPage) {
                items {
                  slug
                  name
                }
                totalItems
                totalPages
              }
            }
        """,
            {"page": page, "perPage": per_page},
        )

    def get_pipelines(self, workspace_slug: str, page: int = 1, per_page: int = 10) -> dict:
        """
        Retrieves a paginated list of pipelines within a specified workspace.

        Parameters:
        ----------
        workspace_slug : str
            The slug identifier of the workspace to retrieve pipelines from.
        page : int, optional
            The page number to retrieve. Defaults to 1.
        per_page : int, optional
            The number of items to retrieve per page. Defaults to 10.

        Returns:
        -------
        dict
            A dictionary containing the list of pipelines (`items`), the total number of items (`totalItems`),
            and the total number of pages (`totalPages`).
        """
        return self.query(
            """
            query workspacePipeline($workspaceSlug: String!,$page:Int, $perPage:Int) {
                pipelines(workspaceSlug: $workspaceSlug, page:$page, perPage:$perPage) {
                    items {
                        id
                        name
                        code
                        type
                    }
                    totalItems
                    totalPages
                }
            }
        """,
            {"workspaceSlug": workspace_slug, "page": page, "perPage": per_page},
        )

    def get_pipeline(self, workspace_slug: str, code: str) -> dict:
        """
        Retrieves details of a specific pipeline by its code within a given workspace.

        Parameters:
        ----------
        workspace_slug : str
            The slug identifier of the workspace containing the pipeline.
        code : str
            The unique code of the pipeline to retrieve.

        Returns:
        -------
        dict
            A dictionary containing the pipeline details.

        Raises:
        ------
        NotFound
            If the pipeline with the specified code is not found in the given workspace.
        """
        pipeline = self.query(
            """
            query getPipeline($workspaceSlug:String!, $pipelineCode:String!){
                pipelineByCode(workspaceSlug: $workspaceSlug, code: $pipelineCode) {
                    id
                    code
                    name
                    currentVersion {
                        id
                    }
                    createdAt
                    updatedAt
                    type
                }
            }
        """,
            {"workspaceSlug": workspace_slug, "pipelineCode": code},
        )
        if not pipeline:
            raise NotFound(f"Pipeline {code} not found in workspace {workspace_slug}.")

        return pipeline

    def run_pipeline(
        self,
        id: uuid.UUID,
        config: typing.Optional[dict] = {},
        version_id: typing.Optional[uuid.UUID] = None,
        send_notification: typing.Optional[bool] = False,
    ) -> dict:
        """
        Executes a specified pipeline with optional configuration, version, and notification settings.

        Parameters:
        ----------
        id : str
            The unique identifier of the pipeline to run.
        config : dict, optional
            A dictionary of configuration settings to apply to the pipeline
        version_id : str, optional
            The specific version ID of the pipeline to run.
        send_notification : bool, optional
            Whether to send notification emails upon pipeline execution, false by default.

        Returns:
        -------
        dict
            A dictionary containing details of the pipeline run, including its ID.

        Raises:
        ------
        NotFound
            If the pipeline with the specified ID is not found.
        Exception
            If the pipeline run fails for any other reason, the error message is raised.
        """
        result = self.query(
            """
            mutation RunWorkspacePipeline($input: RunPipelineInput!) {
                runPipeline(input: $input) {
                    success
                    errors
                    run {
                        id
                    }
                }
            }
        """,
            {
                "input": {
                    "id": id,
                    "config": config,
                    "versionId": version_id,
                    "sendMailNotifications": send_notification,
                }
            },
        )

        if not result["runPipeline"]["success"]:
            if "PIPELINE_NOT_FOUND" in result["runPipeline"]["errors"]:
                raise NotFound(f"Pipeline {id} not found.")

            raise Exception(result["runPipeline"]["errors"])

        return result["runPipeline"]["run"]

    def get_pipeline_run(self, run_id: uuid.UUID):
        """
        Retrieves details of a specific pipeline run by its ID.

        Parameters:
        ----------
        run_id : str
            The unique identifier of the pipeline run to retrieve.

        Returns:
        -------
        dict
            A dictionary containing the pipeline run details.

        Raises:
        ------
        NotFound
            If the pipeline run with the specified ID is not found.
        """
        result = self.query(
            """
            query pipelineRun($runId: UUID!) {
                pipelineRun(id: $runId) {
                    id
                    version {
                        name
                    }
                    status
                    config
                    executionDate
                    duration
                    triggerMode
                    pipeline {
                        id
                    }
                }
            }
        """,
            {"runId": run_id},
        )
        if not result["pipelineRun"]:
            raise NotFound(f"Run {run_id}  not found.")

        return result["pipelineRun"]



from collections import namedtuple
from pathlib import Path

import polars as pl
import pytest
import responses

from openhexa.toolbox.dhis2 import DHIS2
from openhexa.toolbox.dhis2.api import DHIS2Error

VERSIONS = ["2.36", "2.37", "2.38", "2.39", "2.40", "2.41"]


@pytest.fixture
def con():
    Connection = namedtuple("Connection", ["url", "username", "password"])
    return Connection("http://localhost:8080", "admin", "district")


@responses.activate
@pytest.mark.parametrize("version", VERSIONS)
def test_data_elements(version, con):
    responses_dir = Path("tests", "dhis2", "responses", version)
    responses._add_from_file(Path(responses_dir, "dhis2_init.yaml"))
    responses._add_from_file(Path(responses_dir, "data_elements.yaml"))
    api = DHIS2(con, cache_dir=None)
    r = api.meta.data_elements()
    df = pl.DataFrame(r)
    assert len(df) > 1000


@responses.activate
@pytest.mark.parametrize("version", VERSIONS)
def test_datasets(version, con):
    responses_dir = Path("tests", "dhis2", "responses", version)
    responses._add_from_file(Path(responses_dir, "dhis2_init.yaml"))
    responses._add_from_file(Path(responses_dir, "datasets.yaml"))
    api = DHIS2(con, cache_dir=None)
    r = api.meta.datasets()
    df = pl.DataFrame(r)
    assert not df.is_empty()


@responses.activate
@pytest.mark.parametrize("version", VERSIONS)
def test_organisation_unit_levels(version, con):
    responses_dir = Path("tests", "dhis2", "responses", version)
    responses._add_from_file(Path(responses_dir, "dhis2_init.yaml"))
    responses._add_from_file(Path(responses_dir, "organisation_unit_levels.yaml"))
    api = DHIS2(con, cache_dir=None)
    r = api.meta.organisation_unit_levels()
    df = pl.DataFrame(r)
    assert not df.is_empty()


@responses.activate
@pytest.mark.parametrize("version", VERSIONS)
def test_organisation_units(version, con):
    responses_dir = Path("tests", "dhis2", "responses", version)
    responses._add_from_file(Path(responses_dir, "dhis2_init.yaml"))
    responses._add_from_file(Path(responses_dir, "organisation_units.yaml"))
    api = DHIS2(con, cache_dir=None)
    r = api.meta.organisation_units()
    df = pl.DataFrame(r)
    assert not df.is_empty()


@responses.activate
@pytest.mark.parametrize("version", VERSIONS)
def test_indicators(version, con):
    responses_dir = Path("tests", "dhis2", "responses", version)
    responses._add_from_file(Path(responses_dir, "dhis2_init.yaml"))
    responses._add_from_file(Path(responses_dir, "indicators.yaml"))
    api = DHIS2(con, cache_dir=None)
    r = api.meta.indicators()
    df = pl.DataFrame(r)
    assert not df.is_empty()


@responses.activate
@pytest.mark.parametrize("version", VERSIONS)
def test_organisation_unit_groups(version, con):
    responses_dir = Path("tests", "dhis2", "responses", version)
    responses._add_from_file(Path(responses_dir, "dhis2_init.yaml"))
    responses._add_from_file(Path(responses_dir, "organisation_unit_groups.yaml"))
    api = DHIS2(con, cache_dir=None)
    r = api.meta.organisation_unit_groups()
    df = pl.DataFrame(r)
    assert not df.is_empty()


@responses.activate
@pytest.mark.parametrize("version", VERSIONS)
def test_data_element_groups(version, con):
    responses_dir = Path("tests", "dhis2", "responses", version)
    responses._add_from_file(Path(responses_dir, "dhis2_init.yaml"))
    responses._add_from_file(Path(responses_dir, "data_element_groups.yaml"))
    api = DHIS2(con, cache_dir=None)
    r = api.meta.data_element_groups()
    df = pl.DataFrame(r)
    assert not df.is_empty()


@responses.activate
@pytest.mark.parametrize("version", VERSIONS)
def test_category_option_combos(version, con):
    responses_dir = Path("tests", "dhis2", "responses", version)
    responses._add_from_file(Path(responses_dir, "dhis2_init.yaml"))
    responses._add_from_file(Path(responses_dir, "category_option_combos.yaml"))
    api = DHIS2(con, cache_dir=None)
    r = api.meta.category_option_combos()
    df = pl.DataFrame(r)
    assert not df.is_empty()


@responses.activate
@pytest.mark.parametrize("version", VERSIONS)
def test_indicator_groups(version, con):
    responses_dir = Path("tests", "dhis2", "responses", version)
    responses._add_from_file(Path(responses_dir, "dhis2_init.yaml"))
    responses._add_from_file(Path(responses_dir, "indicator_groups.yaml"))
    api = DHIS2(con, cache_dir=None)
    r = api.meta.indicator_groups()
    df = pl.DataFrame(r)
    assert not df.is_empty()


@responses.activate
@pytest.mark.parametrize("version", VERSIONS)
def test_data_value_sets_get_data_elements(version, con):
    responses_dir = Path("tests", "dhis2", "responses", version)
    responses._add_from_file(Path(responses_dir, "dhis2_init.yaml"))
    responses._add_from_file(Path(responses_dir, "data_value_sets_get_data_elements.yaml"))
    api = DHIS2(con, cache_dir=None)

    if api.version < "2.39":
        with pytest.raises(DHIS2Error):
            r = api.data_value_sets.get(
                data_elements=["fbfJHSPpUQD", "cYeuwXTCPkU"],
                periods=["202306", "202307", "202308"],
                org_units=["uNEhNuBUr0i"],
            )

    else:
        r = api.data_value_sets.get(
            data_elements=["fbfJHSPpUQD", "cYeuwXTCPkU"],
            periods=["202306", "202307", "202308"],
            org_units=["uNEhNuBUr0i"],
        )
        df = pl.DataFrame(r)
        assert not df.is_empty()


@responses.activate
@pytest.mark.parametrize("version", VERSIONS)
def test_data_value_sets_get_dataset(version, con):
    responses_dir = Path("tests", "dhis2", "responses", version)
    responses._add_from_file(Path(responses_dir, "dhis2_init.yaml"))
    responses._add_from_file(Path(responses_dir, "data_value_sets_get_dataset.yaml"))
    api = DHIS2(con, cache_dir=None)

    r = api.data_value_sets.get(
        datasets=["BfMAe6Itzgt"], periods=["202306", "202307", "202308"], org_units=["uNEhNuBUr0i"]
    )
    df = pl.DataFrame(r)
    assert not df.is_empty()


@responses.activate
@pytest.mark.parametrize("version", VERSIONS)
def test_data_value_sets_get_start_end(version, con):
    responses_dir = Path("tests", "dhis2", "responses", version)
    responses._add_from_file(Path(responses_dir, "dhis2_init.yaml"))
    responses._add_from_file(Path(responses_dir, "data_value_sets_get_start_end.yaml"))
    api = DHIS2(con, cache_dir=None)

    r = api.data_value_sets.get(
        datasets=["BfMAe6Itzgt"], start_date="2023-06-01", end_date="2023-09-01", org_units=["uNEhNuBUr0i"]
    )
    df = pl.DataFrame(r)
    assert not df.is_empty()


@responses.activate
@pytest.mark.parametrize("version", VERSIONS)
def test_analytics_get_data_elements(version, con):
    responses_dir = Path("tests", "dhis2", "responses", version)
    responses._add_from_file(Path(responses_dir, "dhis2_init.yaml"))
    responses._add_from_file(Path(responses_dir, "analytics_get_data_elements.yaml"))
    api = DHIS2(con, cache_dir=None)

    r = api.analytics.get(
        data_elements=["fbfJHSPpUQD", "cYeuwXTCPkU"], periods=["202306", "202307", "202308"], org_units=["uNEhNuBUr0i"]
    )
    df = pl.DataFrame(r)
    assert not df.is_empty()


@responses.activate
@pytest.mark.parametrize("version", VERSIONS)
def test_analytics_get_indicators(version, con):
    responses_dir = Path("tests", "dhis2", "responses", version)
    responses._add_from_file(Path(responses_dir, "dhis2_init.yaml"))
    responses._add_from_file(Path(responses_dir, "analytics_get_indicators.yaml"))
    api = DHIS2(con, cache_dir=None)

    r = api.analytics.get(
        indicators=["Uvn6LCg7dVU", "OdiHJayrsKo"],
        periods=["202306", "202307", "202308"],
        org_units=["uNEhNuBUr0i"],
        include_cocs=False,
    )
    df = pl.DataFrame(r)
    assert not df.is_empty()


@responses.activate
@pytest.mark.parametrize("version", VERSIONS)
def test_analytics_get_org_unit_levels(version, con):
    responses_dir = Path("tests", "dhis2", "responses", version)
    responses._add_from_file(Path(responses_dir, "dhis2_init.yaml"))
    responses._add_from_file(Path(responses_dir, "analytics_get_org_unit_levels.yaml"))
    api = DHIS2(con, cache_dir=None)

    r = api.analytics.get(
        data_elements=["fbfJHSPpUQD", "cYeuwXTCPkU"], periods=["202306", "202307", "202308"], org_unit_levels=[2]
    )
    df = pl.DataFrame(r)
    assert not df.is_empty()



openhexa_mocked_workspaces = {
    "workspaces": {
        "items": [
            {"slug": "demo-workspace-14f264", "name": "Demo Workspace"},
            {"slug": "demo-custom-env-11a27f", "name": "With custom env"},
        ],
        "totalItems": 2,
        "totalPages": 1,
    }
}

openhexa_mocked_workspaces_pipelines = {
    "pipelines": {
        "items": [
            {
                "id": "04143a0c-db65-430a-977f-f834cedc8a7c",
                "name": "dataset_param",
                "code": "dataset-param",
                "type": "zipFile",
            },
            {"id": "22261aca-b960-4d4a-b5d9-57c5ca21c405", "name": "lqas", "code": "lqas", "type": "zipFile"},
        ],
        "totalItems": 2,
        "totalPages": 1,
    }
}

openhexa_mocked_pipeline = {
    "pipelineByCode": {
        "id": "04143a0c-db65-430a-977f-f834cedc8a7c",
        "code": "dataset-param",
        "name": "dataset_param",
        "currentVersion": {"id": "8cd863b8-11f3-4020-b7be-a701f339459b"},
        "createdAt": "2024-02-13T09:58:58.370Z",
        "updatedAt": "2024-02-13T09:58:58.370Z",
        "type": "zipFile",
    }
}

openhexa_mocked_pipeline_run = {
    "pipelineRun": {
        "id": "cf47a668-2772-4a60-b226-431933d22640",
        "version": {"name": "4"},
        "status": "queued",
        "config": {},
        "executionDate": "2024-08-09T11:14:26.701Z",
        "duration": 0,
        "triggerMode": "manual",
        "pipeline": {"id": "77d90864-3bcb-42f8-812d-9a894138ac80"},
    }
}
